/* $Id: slideshow.c,v 1.7 2007-12-29 00:54:28 chris Exp $
 *
 * Olivetti M20 slide show program
 *
 * Written by Christian Groessler.
 *
 * This program is in the public domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/pcos.h>

#define DEFAULT_DELAY 20    /* seconds */
#define MAX_NUM_OF_BMPS 256

struct bmp_info {
	unsigned long offset; /* bfOffBits */
	unsigned int width;   /* biWidth */
	unsigned int height;  /* biHeight */
};

#define FALSE 0
#define TRUE 1
#define SCREENBUFSIZE 16384
int bmps, cur_bmp;
int delay_secs = DEFAULT_DELAY;
int single;  /* 1 - quit after last picture, 0 - endless loop */
int effect;  /* 1 - use effect */
char bmp_files[MAX_NUM_OF_BMPS][32];  /* file names of the bmp files */
struct bmp_info file_info[MAX_NUM_OF_BMPS];  /* cache image data from file format check */
char search_name[] = "*.bmp";
char search_name2[] = "*.bm";
char name_buf[32];
char bmp_hdr[64];

unsigned short *screen = (unsigned short *)0x3000000;  /* segment #3 */
void make_effect(unsigned short *screen_buf);

void my_cls(void)
{
	memset(screen, 0, SCREENBUFSIZE);
	_pcos_cls();
}

void banner(void)
{
	printf("slideshow 0.9 -- M20 slide show program for monochrome BMP files\n");
	printf("Written by Christian Groessler, 2007. Put into the public domain.\n\n");
}

/* check file for correct bmp format (resolution, etc.) */

int bmp_format_ok(char *filename)
{
	size_t numr;
	unsigned long width, height;
	FILE *f = fopen(filename, "rb");
	if (! f) return 0;

	/* read the 1st 64 bytes and look at them */
	numr = fread(bmp_hdr, 64, 1, f);
	fclose(f);
	if (! numr) return 0;

	/* check header */
	if (bmp_hdr[0] != 'B' || bmp_hdr[1] != 'M')
		return 0;

	/* check planes */
	if ((bmp_hdr[26] | (bmp_hdr[27] << 8)) != 1)
		return 0;

	/* check bits per pixel */
	if ((bmp_hdr[28] | (bmp_hdr[29] << 8)) != 1)
		return 0;

	/* check resolution */
	width = bmp_hdr[18] | (bmp_hdr[19] << 8) | ((unsigned long)bmp_hdr[20] << 16) | ((unsigned long)bmp_hdr[21] << 24);
	height = bmp_hdr[22] | (bmp_hdr[23] << 8) | ((unsigned long)bmp_hdr[24] << 16) | ((unsigned long)bmp_hdr[25] << 24);

	if (width > 512 || height > 256) return 0;
	file_info[bmps].width = (unsigned int)width;
	file_info[bmps].height = (unsigned int)height;

	/* now variable 'width' is misused to get offset */
	width = bmp_hdr[10] | (bmp_hdr[11] << 8) | ((unsigned long)bmp_hdr[12] << 16) | ((unsigned long)bmp_hdr[13] << 24);
	file_info[bmps].offset = width;

	fclose(f);
	return 1;
}

int keyboard_event(unsigned char *key, int *status)
{
	int st;
	unsigned char st1;
	if (_pcos_lookbyte(DID_CONSOLE, key, &st1))
		return 1;	/* error from LookByte */
	else {
		if (st1) { /* key was pressed */
			st = _pcos_getbyte(DID_CONSOLE, key);
			if (st)  /* error from GetByte */
				*status = st;
			else
				*status = PCOS_ERR_OK;
			return 1;
		}
	}
	return 0;  /* nothing happened */
}

#ifdef C_EFFECT
void make_effect(unsigned short *screen_buf)
{
	int i, j;
	unsigned short mask1, mask2;

	for (i = 0; i < 16; i++) { /* bit-# */
		mask1 = 1 << i;
		mask2 = ~mask1;
		for (j = 0; j < SCREENBUFSIZE / 2; j++) {
			*(screen + j) = (*(screen + j) & mask2) | (*(screen_buf + j) & mask1);
		}
	}
}
#endif /* #ifdef C_EFFECT */

void slideshow(void)
{
	int st;
	unsigned int ix, iy, ep1, ep2, height, width, xadj, skipping = 0;
	unsigned char key;
	unsigned short *screen_buf;
	FILE *fin;
	struct timeval starttime, curtime;

	/* allocate buffer to prepare image in */
	screen_buf = malloc(SCREENBUFSIZE);
	if (! screen_buf) {
		printf("cannot allocate %d bytes: %s\n", SCREENBUFSIZE, strerror(errno));
		return;
	}

	while (1) {
		gettimeofday(&starttime, NULL);  /* remember time */
		if (skipping) {
			skipping = 0;
			printf("skipping to %s\n", bmp_files[cur_bmp]);
		}
		memset(screen_buf, 0, SCREENBUFSIZE);

		/* open picture file */
		fin = fopen(bmp_files[cur_bmp], "rb");
		if (! fin) {
			free(screen_buf);
			my_cls();  /* clear screen */
			printf("cannot open '%s': %s\n", bmp_files[cur_bmp], strerror(errno));
			return;
		}

		st = fseek(fin, file_info[cur_bmp].offset, SEEK_SET);
		if (st) {
			free(screen_buf);
			my_cls();  /* clear screen */
			printf("fseek returned error: %s\n", strerror(errno));
			return;
		}

		/* create the screen image */
		height = file_info[cur_bmp].height;
		width = file_info[cur_bmp].width;
		iy = (255 - (256 - height)) * 32;
		xadj = (256 - width / 2) / 16;
		while (height--) {
			for (ix = 0; ix < width / 16; ix++) {
				ep1 = getc(fin);
				ep2 = getc(fin);
				*(screen_buf + iy + xadj + ix) = ep2 | (ep1 << 8);
				if (keyboard_event(&key, &st)) {
					if (st != PCOS_ERR_OK) {
						free(screen_buf);
						my_cls();
						printf("ERROR: %u\n", st);
						return;
					}
					if (key == 0xed || key == '\r') {
						free(screen_buf);
						my_cls();
						printf("Aborted by user\n");
						return;
					}
					if (key == ' ') {
						fclose(fin);
						my_cls();
						skipping = 1;
						goto next;
					}

				}
			}
			iy -= 32;
		}

                if (effect)
			make_effect(screen_buf);
		else
			memcpy(screen, screen_buf, SCREENBUFSIZE);  /* display picture */

		fclose(fin);

		while (1) {
			gettimeofday(&curtime, NULL);
			if (keyboard_event(&key, &st)) {
				if (st != PCOS_ERR_OK) {
					free(screen_buf);
					my_cls();
					printf("ERROR: %u\n", st);
					return;
				}
				if (key == 0xed || key == '\r') {
					free(screen_buf);
					my_cls();
					printf("Aborted by user\n");
					return;
				}
				if (key == ' ') break;
			}
			if (curtime.tv_sec - starttime.tv_sec >= delay_secs)
				break;
		}

	next:
		if (++cur_bmp >= bmps) {
			if (single) break;
			cur_bmp = 0;
		}
	}

	free(screen_buf);
	my_cls();  /* clear screen */
}

/* scan for matching bmp files and fill them in bmp_list[] */

void scan_for_bmps(void)
{
	int retval;
	int length;  /* lenght of the file name */
	char *file_pointer;
	int search_mode = 1;  /* search from beginning */
	int drive = -1;	 /* default drive */

	printf("scanning for picture files....\n");

	/* search for *.bmp files */
	while (1) {
		length = sizeof(search_name) - 1;
		file_pointer = name_buf;
		retval = _pcos_search(drive, search_mode, &length, &file_pointer, search_name);
		if (retval != PCOS_ERR_OK) break;
		search_mode = 0;   /* from now on search from the last file found */
		name_buf[length] = 0;  /* zero terminate name */
		printf("found %s -- checking format...", name_buf);
		if (bmp_format_ok(name_buf)) {
			printf(" ok  (%ux%u)\n", file_info[bmps].width, file_info[bmps].height);
			strcpy(bmp_files[bmps++], name_buf);
		}
		else {
			printf(" invalid format, IGNORED\n");
		}
	}
	if (retval != PCOS_ERR_NOENT) {
		printf("_pcos_search returned error %d\n", retval);
		bmps = 0;
		return;
	}

	/* search for *.bm files (don't ask why :-) */
	search_mode = 1;
	while (1) {
		length = sizeof(search_name2) - 1;
		file_pointer = name_buf;
		retval = _pcos_search(drive, search_mode, &length, &file_pointer, search_name2);
		if (retval != PCOS_ERR_OK) break;
		search_mode = 0;   /* from now on search from the last file found */
		name_buf[length] = 0;  /* zero terminate name */
		printf("found %s -- checking format...", name_buf);
		if (bmp_format_ok(name_buf)) {
			printf(" ok  (%ux%u)\n", file_info[bmps].width, file_info[bmps].height);
			strcpy(bmp_files[bmps++], name_buf);
		}
		else {
			printf(" invalid format, IGNORED\n");
		}
	}
	if (retval != PCOS_ERR_NOENT) {
		printf("_pcos_search returned error %d\n", retval);
		bmps = 0;
		return;
	}
}

static void command_line(int argc, char **argv)
{
	int i = 0;
	int delay_mode = FALSE;
	char *err_arg;
	unsigned long delayval;

	argc--;
	while(i++, argc--) {
		if (delay_mode) {
			delayval = strtoul(*(argv + i), &err_arg, 0);
			if (delayval == ULONG_MAX) { /* overflow */
				printf("command line argument for %%d: %s\n", strerror(errno));
				exit(1);
			}
			if (*err_arg || delayval > 7200) {
				printf("invalid argument for %%d\n");
				exit(2);
			}
			delay_mode = FALSE;
			delay_secs = delayval;
			continue;
		}
		if (! strcmp(*(argv+i), "%d")) {
			delay_mode = TRUE;
			err_arg = "%%d";
			continue;
		}
		if (! strcmp(*(argv+i), "%s")) {
			single = 1;
			continue;
		}
		if (! strcmp(*(argv+i), "%e")) {
			effect = 1;
			continue;
		}
		/* unknown switch */
		banner();
		printf("Usage:\tslideshow <%%d secs> <%%s> <%%e>\n");
		printf("\t%%s - stop after all pictures have been displayed once\n");
		printf("\t%%d - number of seconds to wait between slides\n");
		printf("\t%%e - switch slides with an `effect'\n");
		exit(1);
	}
	if (delay_mode) {
		printf("%s needs a parameter\n", err_arg);
		exit(1);
	}
}


/* entry point */

int main(int argc, char **argv)
{
	command_line(argc, argv);

	banner();

	scan_for_bmps();  /* scan directory for .bmp files */

	if (! bmps) { /* none found */
		printf("no pictures found, terminating...\n");
		return 0;
	}

        printf("\n\n\n\n\
                              SLIDESHOW LOADING...\n\
                             ======================\n\
\n\
                     Press   SPACE     to skip to next slide\n\
                             RETURN    to exit\n\
                             COMMAND-1 to exit\n\n\n\n\n\n\n");

	slideshow();

	return 0;
}

